'// IPAD-OS Time - sets the local PC clock to the time found on the IPAD-OS mail server
'// Created 2006 by Casey Sprangel in PowerBASIC/CC 5.0 (www.powerbasic.com)
'// Some date routines converted to PBDLL/PBCC by Scott Slater
'//============================================================================
#COMPILE EXE                                                  '// create stand-alone program
#OPTIMIZE SIZE                                                '// make a smaller program
#DIM ALL                                                      '// force declare of all variables before use
#TOOLS OFF                                                    '// disable development tools (for smaller code)
#INCLUDE "WIN32API.INC"                                       '// include windows API functions
%myver = 20101022

'//----------------------------------------------------------------------------
'// Structures
TYPE DateSplit                                                '// used for DateDiff
  LowPart AS DWORD
  HighPart AS DWORD
END TYPE

UNION DateQuad                                                '// used for DateDiff interval
  U AS DateSplit
  QuadPart AS QUAD
END UNION

'//----------------------------------------------------------------------------
'// Static values
%HWND_BROADCAST     = &HFFFFFFFF???                           '// flag to send the following message to everyone
%WM_TIMECHANGE      = &H1E                                    '// flag to tell others that the time is changed
%TIME_ONE_SECOND    = 10000000&&                              '// used for TimeDate_Change
%TIME_ONE_MINUTE    = 600000000&&                             '// used for TimeDate_Change
%TIME_ONE_HOUR      = 36000000000&&                           '// used for TimeDate_Change
%DATE_ONE_DAY       = 864000000000&&                          '// used for TimeDate_Change
%INT_SECONDS        = 10000000&&                              '// used for DateDiff interval
%INT_MINUTES        = 600000000&&                             '// used for DateDiff interval
%INT_HOURS          = 36000000000&&                           '// used for DateDiff interval
%INT_DAYS           = 864000000000&&                          '// used for DateDiff interval

'//----------------------------------------------------------------------------

DECLARE FUNCTION DateDiff(BYVAL Interval AS QUAD,BYVAL udtSource AS SYSTEMTIME,BYVAL udtDest AS SYSTEMTIME) AS QUAD
DECLARE FUNCTION DaysToDate(BYVAL NumDays AS DWORD) AS STRING
DECLARE FUNCTION Leap(BYVAL Year AS LONG) AS LONG
DECLARE FUNCTION MakeDate(BYVAL Month AS LONG, BYVAL Day AS LONG, BYVAL Year AS LONG) AS STRING
DECLARE FUNCTION MakeTime(Hour AS LONG, Minute AS LONG, Second AS LONG) AS STRING
DECLARE FUNCTION SecondsToTime(BYVAL Secs AS LONG) AS STRING
DECLARE FUNCTION UpdateSystemTime(buf AS STRING) AS LONG
DECLARE SUB TimeDate_Change(BYREF st AS SYSTEMTIME, BYVAL nValue AS QUAD)

'//------------------------------------------------------------------------------
'// MakeDate - Combine the component parts of a date into a displayable string in U.S. format.
'// Note:  All other date routines call this routine to combine the component parts.
FUNCTION MakeDate(BYVAL Month AS LONG, BYVAL Day AS LONG, BYVAL Year AS LONG) AS STRING
  IF Year < 80 THEN                                           '// assume we've rolled into the next century
    Year = Year + 2000
  ELSEIF Year < 100 THEN                                      '// between 1980 and 1999
    Year = Year + 1900
  END IF
  FUNCTION = FORMAT$(Month, "00") + "-" + FORMAT$(Day, "00") + "-" + FORMAT$(Year, "0000")
END FUNCTION

'//------------------------------------------------------------------------------
'// Leap - Return true (1) if specified year is a leap year.
'// Year = year to check
FUNCTION Leap(BYVAL Year AS LONG) AS LONG
  IF Year MOD 4 = 0 THEN                                      '// every four years is a leap year
    FUNCTION = 1                                              '// this is a leap year
    IF Year MOD 100 = 0 THEN                                  '// every hundred years is not a leap year
      IF Year MOD 400 <> 0 THEN                               '// except for every four hundredth year
        FUNCTION = 0                                          '// this is not a leap year
      END IF
    END IF
  END IF
END FUNCTION

'//------------------------------------------------------------------------------
'// DaysToDate - Convert the number of days since 01-01-1900 to a date string.
'// NumDays = number of days since 1900
FUNCTION DaysToDate(BYVAL NumDays AS DWORD) AS STRING
  LOCAL Temp AS LONG
  LOCAL LastYear AS LONG
  LOCAL LeapDays AS LONG
  LOCAL x AS LONG
  LOCAL p AS BYTE PTR
  LOCAL DaysPerMonth AS STRING
  LOCAL Year AS LONG
  LOCAL Month AS LONG
  LOCAL Day AS LONG

  Temp = NumDays + 693595                                     '// Adjust back to number of days since 1/1/0001
  LastYear = Temp \ 365                                       '// turn days into years
  LeapDays = LastYear \ 4 - LastYear \ 100 + LastYear \ 400   '// get leap year interval
  Day = Temp MOD 365 - LeapDays                               '// get day of year
  DO WHILE Day < 1                                            '// Day is negative.  Correct that by backing up a year at a time until Day > 0.
    Day = Day + 365 + Leap(LastYear)                          '// add a year of days
    DECR LastYear                                             '// go to previous year
  LOOP
  Year = LastYear + 1                                         '// get year
  x = 28 + Leap(Year)                                         '// get number of February days in the year
  DaysPerMonth = CHR$(31,x,31,30,31,30,31,31,30,31,30,31)     '// set up a table of days in each month and access it with a pointer
  Month = 1                                                   '// default month to January
  p = STRPTR(DaysPerMonth)                                    '// set pointer to our table
  DO WHILE Day > @p                                           '// loop while day is more than days in month of value at our pointer
    Day = Day - @p                                            '// take away a month of days
    INCR Month                                                '// move up a month
    INCR p                                                    '// move pointer to next table element
  LOOP
  FUNCTION = MakeDate(Month, Day, Year)                       '// return formatted date
END FUNCTION

'//------------------------------------------------------------------------------
'// MakeTime - Combine the component parts of a time into a displayable string in 24 hour U.S. format.
'// Note:  All other time routines call this routine to combine their component parts.
FUNCTION MakeTime(Hour AS LONG, Minute AS LONG, Second AS LONG) AS STRING
  FUNCTION = FORMAT$(Hour, "00") + ":" + FORMAT$(Minute, "00") + ":" + FORMAT$(Second, "00")
END FUNCTION

'//------------------------------------------------------------------------------
'// SecondsToTime$ - Return a time string for a given number of seconds since midnight.
'// Secs = seconds to convert into a time string
FUNCTION SecondsToTime(BYVAL Secs AS LONG) AS STRING
  LOCAL Hour   AS LONG
  LOCAL Minute AS LONG
  LOCAL Second AS LONG

  Second = Secs MOD 60                                        '// get number of seconds (the part not evenly divisible by 60)
  Secs = Secs \ 60                                            '// turn into minutes
  Minute = Secs MOD 60                                        '// get number of minutes (the part not evenly divisible by 60)
  Hour = Secs \ 60                                            '// turn into hours
  FUNCTION = MakeTime(Hour, Minute, Second)                   '// return formatted time
END FUNCTION

'//------------------------------------------------------------------------------
'// Return the difference between two times granular to a given time interval
FUNCTION DateDiff(BYVAL Interval AS QUAD,BYVAL udtSource AS SYSTEMTIME,BYVAL udtDest AS SYSTEMTIME) AS QUAD
  LOCAL udtFileSource AS FILETIME
  LOCAL udtFileDest AS FILETIME
  LOCAL udtQuadSource AS DateQuad
  LOCAL udtQuadDest AS DateQuad

  SystemTimetoFileTime udtSource, udtFileSource               '// Convert to filetime
  SystemTimetoFileTime udtDest, udtFileDest
  udtQuadSource.u.LowPart = udtFileSource.dwLowDateTime       '// Convert to quad time
  udtQuadSource.u.HighPart = udtFileSource.dwHighDateTime
  udtQuadDest.u.LowPart = udtFileDest.dwLowDateTime
  udtQuadDest.u.HighPart = udtFileDest.dwHighDateTime
  FUNCTION = (udtQuadDest.QuadPart - udtQuadSource.QuadPart) \ interval '// return answer
END FUNCTION

'//----------------------------------------------------------------------------
'// Add or subtract values from a SYSTEMTIME structure
SUB TimeDate_Change(BYREF st AS SYSTEMTIME, BYVAL nValue AS QUAD)
  LOCAL ft AS FILETIME
  LOCAL p64 AS QUAD PTR

  IF st.wYear = 0 THEN                                        '// user did not pass a date?
    GetLocalTime st                                           '// Get current time if none specified
  END IF
  SystemTimeToFileTime st, ft                                 '// Convert to a FILETIME
  p64 = VARPTR(ft)                                            '// Change the date using a pointer
  @p64 = @p64 + nValue                                        '// add the offset
  FileTimeToSystemTime ft, st                                 '// Convert back
END SUB

'//----------------------------------------------------------------------------
'// Original time set code from Mike Doty
FUNCTION UpdateSystemTime(buf AS STRING) AS LONG              '// buf must be in "yyyymmdd hh:mm:ss" format
  LOCAL v AS SYSTEMTIME                                       '// system time structure
  LOCAL s AS STRING                                           '// working string for easier parsing

  IF INSTR(buf, $SPC) = 0 OR LEN(buf) <> 17 THEN              '// not what we expected? (does not contain space or is not 17 characters long)
    FUNCTION = 1                                              '// set the error flag
    EXIT FUNCTION                                             '// leave so we don't cause trouble
  END IF
  s = LEFT$(buf, 8)                                           '// get the date part of the string
  v.wYear = VAL(LEFT$(s, 4))                                  '// get the year
  v.wMonth = VAL(MID$(s, 5, 2))                               '// get the month
  v.wDay = VAL(RIGHT$(s, 2))                                  '// get the day
  s = RIGHT$(buf, 8)                                          '// get the time part of the string
  v.wHour = VAL(LEFT$(s, 2))                                  '// get the hour
  v.wMinute = VAL(MID$(s, 4, 2))                              '// get the minute
  v.wSecond = VAL(RIGHT$(s, 2))                               '// get the second
  SetSystemTime v                                             '// call the time set function from Windows
  SendMessage %HWND_BROADCAST, %WM_TIMECHANGE, 0&, 0&         '// put it on the system message pump queue
END FUNCTION

'//----------------------------------------------------------------------------
FUNCTION logfile(msg AS STRING) AS LONG                       '// write information to the error log
  LOCAL h AS LONG                                             '// file handle

  STDOUT msg                                                  '// show message on screen too
  h = FREEFILE                                                '// get next available file handle
  OPEN "ITIME.ERR" FOR APPEND AS h                            '// append to our log file
  PRINT #h, DATE$ + $SPC + TIME$ + ": " + msg                 '// write message to file with date and time stamp
  CLOSE h
END FUNCTION

'//----------------------------------------------------------------------------
FUNCTION PBMAIN() AS LONG                                     '// Program starts here
  DIM buf AS STRING                                           '// main working string buffer
  DIM changetime AS LONG                                      '// flag for setting time (set from command line)
  DIM hTCP AS LONG                                            '// handle for tcp communication
  DIM i AS LONG                                               '// main working counter/pointer
  DIM ipad AS STRING                                          '// the IPAD server address
  DIM lt AS SYSTEMTIME                                        '// holds the local workstation time
  DIM myTZ AS LONG                                            '// workstation time zone in number format
  DIM myport AS LONG                                          '// the SMTP port to the IPAD-OS mail server
  DIM qdDifference AS QUAD                                    '// holds the difference between TZ and LZ
  DIM st AS SYSTEMTIME                                        '// holds the IPAD-OS time
  DIM tz AS LONG                                              '// IPAD-OS time zone in number format
  DIM tzi AS TIME_ZONE_INFORMATION                            '// used to find time zone offset from GMT
  DIM gmtdate AS STRING
  DIM cdate AS STRING
  DIM gmttime AS STRING
  DIM ctime AS STRING
  DIM nettime AS DWORD

  buf = "IPAD-OS TIME " + FORMAT$(%myver)                     '// program title and version
  CONSOLE NAME buf                                            '// change the console title bar to our title
  STDOUT buf                                                  '// show the title on the console screen
  buf = COMMAND$                                              '// get the command line from the user
  IF INSTR(buf, "/?") > 0 OR buf = "" THEN                    '// did user ask for help?
    STDOUT "This program gets the time and date from any IPAD-OS mail server and sets the"
    STDOUT "local system clock. This works only if the time zone is the same for both the"
    STDOUT "IPAD-OS server and this workstation."
    STDOUT ""
    STDOUT "Syntax: ITIME ip_address[:port] [/SET]"
    STDOUT "  The IP address may include an optional TCP port. 587 is default."
    STDOUT "  The optional /SET switch forces an update to the system clock."
    EXIT FUNCTION                                             '// leave the program, we are done
  END IF
  i = INSTR(UCASE$(buf), "/SET")                              '// locate the command line switch
  IF i > 0 THEN                                               '// did we find an offset?
    ChangeTime = -1                                           '// set flag to true
    buf = TRIM$(LEFT$(buf, i - 1) + MID$(buf, i + 4))         '// remove switch from the command line
  END IF
  IF INSTR(buf, ANY CHR$(0 TO 45,47,59 TO 255 )) > 0 THEN     '// are there any bad characters in our address?
    STDOUT "Syntax error. Use /? on command line for help."
    FUNCTION = 1                                              '// set errorlevel 1
    EXIT FUNCTION                                             '// leave the program
  END IF
  ipad = buf                                                  '// make copy because we may change
  i = INSTR(ipad, ":")                                        '// locate colon in the address
  IF i > 0 THEN                                               '// is port number specified?
    myport = ABS(VAL(MID$(ipad, i + 1)))                      '// grab port info off of end of string
    ipad = LEFT$(ipad, i - 1)                                 '// remove port info from main string
  ELSE                                                        '// no port given?
    myport = 587                                              '// default to SMTP port (587) if none is explicitly given
  END IF
  IF myport < 1 OR myport > 65535 THEN                        '// port outside valid range?
    STDOUT "Error: Port number must be in the range 1 to 65535."
    FUNCTION = 1                                              '// set errorlevel 1
    EXIT FUNCTION                                             '// leave the program
  END IF
  STDOUT "Opening connection to " + ipad + " (port " + FORMAT$(myport) + ")" '// tell users what we are doing
  hTCP = FREEFILE                                             '// get next available handle
  TCP OPEN PORT myport AT ipad AS #hTCP TIMEOUT 300000        '// timeout is in milliseconds
  IF ERR THEN                                                 '// was there a problem opening the connection?
    logfile "Error connecting to " + ipad + " (Err:" + FORMAT$(ERR) + ")"
    FUNCTION = 1                                              '// set errorlevel 1
    EXIT FUNCTION                                             '// leave program
  END IF                                                      '// was there a problem opening the connection?
  IF NOT EOF(hTCP) THEN                                       '// is there something in the inbound buffer?
    TCP LINE #hTCP, buf                                   '// get the connection message
    TCP PRINT #hTCP, "QUIT" + $CR                             '// be polite to the server before we drop the connection
  END IF
  TCP CLOSE #hTCP                                             '// close the connection, we are done
  IF TRIM$(buf) = "" THEN                                 '// no answer?
    logfile "No reply from " + ipad + ":" + FORMAT$(myport)
    FUNCTION = 1                                              '// reutrn errorlevel 1
    EXIT FUNCTION                                             '// leave program
  END IF
  IF INSTR(buf, "IPAD") = 0 THEN                          '// not an IPAD-OS server?
    logfile "Error: " + ipad + ":" + FORMAT$(myport) + " is not an IPAD-OS mail server!"
    STDOUT "       See -> http://www.OneBoxSolution.com"'    ShellExecute %HWND_DESKTOP, "Open", "http://www.oneboxsolution.com", BYVAL 0, BYVAL 0, %SW_SHOWNORMAL
    logfile "Received: " + buf
    buf = ""                                              '// forget what the server said
    FUNCTION = 1                                              '// reutrn errorlevel 1
    EXIT FUNCTION                                             '// leave program
  END IF
  GetSystemTime lt                                            '// get the local system time
  buf = RIGHT$(buf, 26)                                       '// get just the interesting part of the IPAD-OS answer
  STDOUT "IPAD-OS time is: " + buf                            '// show user the IPAD-OS time
  '// 220 example.com ESMTP (IOA-IPAD #.#) ready at Mon, 10 Apr 2006 19:09:23 -0600
  tz = VAL(RIGHT$(buf, 5))                                    '// get the time zone including the "+" or "-" value
  st.wDay = VAL(LEFT$(buf, 2))                                '// get the day
  SELECT CASE UCASE$(MID$(buf, 4, 3))                         '// change the month name into a number
    CASE "JAN" : st.wMonth = 01
    CASE "FEB" : st.wMonth = 02
    CASE "MAR" : st.wMonth = 03
    CASE "APR" : st.wMonth = 04
    CASE "MAY" : st.wMonth = 05
    CASE "JUN" : st.wMonth = 06
    CASE "JUL" : st.wMonth = 07
    CASE "AUG" : st.wMonth = 08
    CASE "SEP" : st.wMonth = 09
    CASE "OCT" : st.wMonth = 10
    CASE "NOV" : st.wMonth = 11
    CASE ELSE  : st.wMonth = 12
  END SELECT
  st.wYear = VAL(MID$(buf, 8, 4))                             '// get the year
  buf = MID$(buf, 13, 8)                                      '// get the time
  st.wHour = VAL(LEFT$(buf, 2))                               '// pull out hours
  st.wMinute = VAL(MID$(buf, 4, 2))                           '// minutes
  st.wSecond = VAL(RIGHT$(buf, 2))                            '// seconds
  SELECT CASE GetTimeZoneInformation(tzi)                     '// find our workstation time zone settings
  CASE %TIME_ZONE_ID_DAYLIGHT                                 '// we are on daylight savings time
    buf = IIF$((tzi.bias + tzi.DaylightBias) <= 0, "+", "-") _ '// plus for east of GMT, minus for west
          + FORMAT$((tzi.bias + tzi.DaylightBias) \ 60, "00") _ '// hours
          + FORMAT$((tzi.bias + tzi.DaylightBias) MOD 60, "00") '// minutes
  CASE %TIME_ZONE_ID_STANDARD                                 '// we are on standard time
    buf = IIF$((tzi.bias + tzi.StandardBias) <= 0, "+", "-") _ '// plus for east of GMT, minus for west
          + FORMAT$((tzi.bias + tzi.StandardBias) \ 60, "00") _ '// hours
          + FORMAT$((tzi.bias + tzi.StandardBias) MOD 60, "00") '// minutes
  CASE ELSE                                                   '// no time zone info?
    buf = "-0000"                                             '// no offset
  END SELECT
  myTZ = VAL(buf)                                             '// turn into a number so we can do math
  STDOUT "Local system time zone offset: " + buf
  IF tz\100 <> tz/100 THEN                                    '// not an even number of hours offset?
    STDOUT "Error: PEI time zone is not supported."           '// a Canadian thing?
    FUNCTION = 1                                              '// set errorlevel 1
    EXIT FUNCTION                                             '// leave, we'll fix this later
  END IF
  TimeDate_Change st, %TIME_ONE_HOUR * (-tz / 100)            '// apply the timezone offset
  buf = FORMAT$(st.wYear, "0000") + FORMAT$(st.wMonth, "00") + FORMAT$(st.wDay,"00") + $SPC + _
             FORMAT$(st.wHour, "00") + ":" + FORMAT$(st.wMinute, "00") + ":" + FORMAT$(st.wSecond, "00")
  qdDifference = ABS(DateDiff(%INT_SECONDS, st, lt))          '// find the difference between IPAD-OS and local time
  STDOUT "GMT: " + buf                                        '// show GMT
  IF qdDifference > 0 THEN                                    '// is system time different from the IPAD-OS time?
    IF changetime THEN                                        '// should we change the workstation clock?
      FUNCTION = UpdateSystemTime(buf)                        '// set the time and return success or failure as ERRORLEVEL
    END IF
    INCR qdDifference                                         '// add one for the zero base
    STDOUT "LOC: " + FORMAT$(lt.wYear, "0000") + FORMAT$(lt.wMonth, "00") + FORMAT$(lt.wDay,"00") + $SPC + _
         FORMAT$(lt.wHour, "00") + ":" + FORMAT$(lt.wMinute, "00") + ":" + FORMAT$(lt.wSecond, "00")
    IF changetime THEN                                        '// should we change the workstation clock?
      STDOUT "Workstation clock was " + FORMAT$(qdDifference) + " second" + IIF$(qdDifference = 1,"","s") + " off."
    ELSE                                                      '// do not change the workstation clock
      STDOUT "Workstation clock is " + FORMAT$(qdDifference) + " second" + IIF$(qdDifference = 1,"","s") + " off."
      STDOUT "System clock was not changed."                  '// tell user we did not change anything
    END IF
  ELSE                                                        '// the clocks match!
    STDOUT "System clock is correct."
  END IF
END FUNCTION
